
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "ciao.h"
#include "system.h"
#include "dir.h"

#ifdef WIN32
#pragma warning(disable : 4996)
#endif

#define PATH_EXCEEDS_THRESHOLD 	"PATH_EXCEEDS_THRESHOLD"
#define PATH_BELOW_THRESHOLD	"PATH_BELOW_THRESHOLD"
#define PATH_EXCEEDS_COUNT	"PATH_EXCEEDS_COUNT"
#define PATH_BELOW_COUNT	"PATH_BELOW_COUNT"
#define PATH_UNAVAILABLE	"PATH_UNAVAILABLE"
#define INVALID_EXPRESSION	"INVALID_EXPRESSION"

#define MAX_PATHS 256

typedef struct {
	int count;
	int64 size;
} pathsize;

static pathsize getsize(char *f, int recursive);


static char *sizestring(int64 size)
{
	static char units[] = {'b', 'k', 'm', 'g'};
	static char ss[10][100];
	static int maxunit = 3;
	static int flipflop = 0;
	int unit = 0;
	
	char *s = ss[flipflop++%10];

	while (size>=10000 && unit<maxunit)
	{
		unit++;
		size/=1024L;
	}
	sprintf(s, "%lld%c", size, units[unit]);
	return s;
}

static pathsize getdirsize(char *f, int recursive)
{
	pathsize result = {0,0L};
	dirhandle *dir = opendirhandle(f);
	if (dir)
	{
		fileinfo fi;
		while (nextfile(dir, &fi))
		{
			if (fi.isdir && recursive)
			{
				char buff[1024];
				pathsize inner; 
				sprintf(buff, "%s/%s", f, fi.name);
				inner = getdirsize(buff, 1);
				if (inner.count>0)
				{
					result.count+=inner.count;
					result.size+=inner.size;
				}
			} 
			else 
			{
				result.count++;
				result.size+=fi.size;
			}
		}
		closedirhandle(dir);
	}
	else
	{
		result.count = -1;
	}
	return result;
}

static pathsize getsize(char *f, int recursive)
{
	fileinfo fi;
	pathsize result = {-1, 0L};
	if (getfileinfo(f, &fi))
	{
		if (fi.isdir)
		{
			result =  getdirsize(f, recursive);
		}
		else
		{
			result.count = 1;
			result.size = fi.size;
		}
	}
	return result;
}

static void compare(pathsize pt, char sign, char *ths, char *path, iao_alarm **alarm)
{
	int gt = sign=='>';
	int64 threshold;
	int sizecompare = 1;
	char *alarmid = NULL;


	while (*ths==' ') ths++; /* whitespace */

	threshold = atol(ths);

	while (*ths>='0' && *ths<='9') ths++; /*skip threshold */
	while (*ths==' ') ths++; /* whitespace */

	switch(*ths)
	{
		case 'b':
		case 'B':
			break;	
		case 'K':
		case 'k':
			threshold*=1024;
			break;
		case 'M':
		case 'm':
			threshold*=(1024*1024);
			break;
		case 'g':
		case 'G':
			threshold*=(1024*1024*1024);
			break;
		default:
			sizecompare = 0;
			break;
	}

	if (sizecompare)
	{
		if (gt)
		{
			if (pt.size>threshold)
			{
                		(**alarm).type = 'A';
                		strncpy((**alarm).resource, path, 255);
                		(**alarm).resource[255] = 0;
                		strcpy((**alarm).id,  alarmid=PATH_EXCEEDS_THRESHOLD);
                		(**alarm).severity = 'M';
                		sprintf((**alarm).description, "Size of path content exceeds threshold (%s/%s)", sizestring(pt.size), sizestring(threshold));
                		(*alarm)++;
			}
		}
		else
		{
			if (pt.size<threshold)
			{
                		(**alarm).type = 'A';
                		strncpy((**alarm).resource, path, 255);
                		(**alarm).resource[255] = 0;
                		strcpy((**alarm).id, alarmid=PATH_BELOW_THRESHOLD);
                		(**alarm).severity = 'M';
                		sprintf((**alarm).description, "Size of path content is below the threshold (%s/%s)", sizestring(pt.size), sizestring(threshold));
                		(*alarm)++;
			}
		}
	}
	else
	{
		if (gt)
		{
			if (pt.count>threshold)
			{
                		(**alarm).type = 'A';
                		strncpy((**alarm).resource, path, 255);
                		(**alarm).resource[255] = 0;
                		strcpy((**alarm).id, alarmid=PATH_EXCEEDS_COUNT);
                		(**alarm).severity = 'M';
                		sprintf((**alarm).description, "File count in directory exceeds threshold (%d/%lld)", pt.count, threshold);
                		(*alarm)++;
			}
		}
		else
		{
			if (pt.count<threshold)
			{
                		(**alarm).type = 'A';
                		strncpy((**alarm).resource, path, 255);
                		(**alarm).resource[255] = 0;
                		strcpy((**alarm).id, alarmid=PATH_BELOW_COUNT);
                		(**alarm).severity = 'M';
                		sprintf((**alarm).description, "File count in directory below threshold (%d/%lld)", pt.count, threshold);
                		(*alarm)++;
			}
		}
	}
}


static void evaluate(char *expression, iao_alarm **alarm)
{
	pathsize pt;
	char sign;
	int r;
	char *signp = strchr(expression, '>');

	if (signp==NULL)
	{
		signp = strchr(expression, '<');
		if (signp==NULL)
		{
                	(**alarm).type = 'A';
                	strncpy((**alarm).resource, expression, 255);
                	(**alarm).resource[255] = 0;
                	strcpy((**alarm).id, INVALID_EXPRESSION);
                	(**alarm).severity = 'm';
                	strcpy((**alarm).description, "Failed to parse path expression\nSee help for exact syntax");
                	(*alarm)++;
			return;
		}
	}

	r = expression[strlen(expression)-1]=='r';
	

	sign = *signp;
	*signp = 0;
	
	pt = getsize(expression, r);

	if (pt.count<0)
	{
		(**alarm).type = 'A';
		strncpy((**alarm).resource, expression, 255);
		(**alarm).resource[255] = 0;
		strcpy((**alarm).id, PATH_UNAVAILABLE);
		(**alarm).severity = 'C';
		strcpy((**alarm).description, "Path cannot be accessed");
		(*alarm)++;
	}
	else
	{
		compare(pt, sign, signp+1, expression, alarm);
	}
	*signp = sign;
}

static void  evaluateall(char *expressions, iao_alarm **alarm)
{
	char *p = expressions;
	while ((p=strchr(expressions, '\n'))) {
		*p = 0;
		if (expressions[0]) 
		{
			evaluate(expressions, alarm);	
		}
		*p = '\n';
		expressions = p+1;
	}
	if (expressions[0]) 
	{
		evaluate(expressions, alarm);
	}
}

int checkpaththresholds(char *expressions, iao_alarm *alarms)
{
	iao_alarm *alarmp = alarms;

    if (expressions==NULL)
    {
        return 0;
    }

	evaluateall(expressions, &alarmp);
	return (int) (alarmp - alarms);
}



#ifdef UNITTEST
int main()
{
	char buff[256] = //"/Users/yuvalperlov/ruon/aents>3m,r\n"
		//"/Users/yuvalperlov/ruon/agents>1m,r\n"
		//"/Users/yuvalperlov/ruon/agents$1m,r\n"
		"/Users/yuvalperlov/ruon/agents/serveron<29,r\n";
	iao_alarm alarms[256];
	int count;
	int i;
	
	while (1)
	{
		count = 0;
		checkpaththresholds(buff, alarms, &count);

		for (i=0;i<count;++i)
		{
			printf("resource: %s, id: %s, sev: %c desc: %s\n",
			alarms[i].resource,
			alarms[i].id,
			alarms[i].severity,
			alarms[i].description);
		}
		sleep(10);
		printf("\n\n\n");
	}
}
#endif
